const Constants = require("../Constants");
const EventEmitter = require("events");
const Self = require("../Structures/Self");
const WebSocket = require("ws");
const logger = require("../Utils/Logger");
const GuildChannel = require("../Structures/GuildChannel");
const Guild = require("../Structures/Guild");
const Message = require("../Structures/Message");

class Websocket extends EventEmitter {
    constructor(client) {
        super(client);
        this.client = client;
        this.lastEvent = null;
        this.sessionId = null;
        this.connect();
        this.lastHeartbeatSent = null;
        this.lastHeartbeatAck = null;
        this.expectingClose = false;
    }

    connect() {
        if (this.client.isConnected) throw new Error("Attempted to connect while already connected.");
        
        this.expectingClose = false;
        try {
            this.ws = new WebSocket(this.client.gwUrl);
            this.client.isConnected = true;
            this.client.ws = this;
            this.registerEventListeners();

            if (this.client.DEBUG) logger.debug("Connected to Litecord gateway!")
        } catch (e) {
            logger.error(`Litecord Gateway Connection Failed!: ${e}`);
        }
    }

    send(op, d) {
        this.ws.send(JSON.stringify({ op, d }));
    }

    identify() {
        return {
            "token": this.client.token,
            "properties": {
                "$properties": process.platform,
                "$browser": "Litecord.js",
                "$device": "Litecord.js"
            },
            "compress": false,
            "large_threshold": 250,
            "presence": {
                "game": this.client.game ? { name: this.client.game, type: 0 } : null,
                "status": this.client.STATUS,
                "since": Date.now(),
                "afk": Boolean(this.client.afk)
            }
        };
    }

    resume() {
        return {
            token: this.client.token,
            session_id: this.sessionId,
            seq: this.lastEvent || null
        };
    }

    singleHeartbeat() {
        this.lastHeartbeatSent = Date.now();
        this.send(Constants.gatewayOpCodes.HEARTBEAT, this.lastEvent);
    }

    heartbeat(interval) {
        if (!this.client.isConnected) return;
        
        let x = setInterval(() => {
            try {
                this.send(Constants.gatewayOpCodes.HEARTBEAT, this.lastEvent);
                this.lastHeartbeatSent = Date.now();
                
                if (this.client.DEBUG) logger.debug(`Sending heartbeat at sequence ${this.lastEvent}`);
            } catch (e) {
                clearInterval(x);
            }
        }, interval - 2000);
    }

    checkIfReady() {
        let unavailableGuilds = 0;

        for (const x of this.client.guilds.values()) {
            if (x.unavailable) unavailableGuilds++;
        }

        if (unavailableGuilds === 0 && this.client.self !== null) return true;

        return false;
    }

    triggerReady() {
        if (this.checkIfReady()) {
            this.client.emit("ready");
        } else {
            if (this.client.DEBUG) logger.debug("Unable to trigger ready, client not ready.")

            return false;
        }
    }

    registerEventListeners() {
        this.ws.on("open", () => {
            if (this.sessionId) {
                this.send(Constants.gatewayOpCodes.RESUME, this.resume());
                this.client.emit("resume");

                if (this.client.DEBUG) logger.debug(`Resumed with Session ID: ${this.sessionId} and token: ${this.client.token}`);
            } else {
                this.send(Constants.gatewayOpCodes.IDENTIFY, this.identify());

                if (this.client.DEBUG) logger.debug(`Authenticated using token: ${this.client.token}`);
            }
        });

        this.ws.on("message", (gatewayMessage) => {
            let res = JSON.parse(gatewayMessage);
            let [ op, type, seq, data ] = [ res.op, res.t, res.s, res.d ];

            if (seq) this.lastEvent = seq;

            this.client.emit("ANY", op, type, data);

            if (op === Constants.gatewayOpCodes.HELLO && data.heartbeat_interval) {
                if (this.client.DEBUG) logger.debug(`Hello triggered, interval: ${data.heartbeat_interval}ms`);
                
                this.heartbeat(data.heartbeat_interval);
            }

            if (op === Constants.gatewayOpCodes.EVENT_DISPATCH && type === "READY" && data.session_id) {
                this.sessionId = data.session_id;
                
                if (this.client.DEBUG) logger.debug(`Ready! Session ID: ${data.session_id}`);
                
                this.client.self = new Self(this.client, data);
                this.singleHeartbeat();
                this.triggerReady();
            }

            if (op === Constants.gatewayOpCodes.HEARTBEAT_ACK) {
                if (this.client.DEBUG) logger.debug("Heartbeat Ack!");
                this.lastHeartbeatAck = Date.now();
            }

            if (op === Constants.gatewayOpCodes.EVENT_DISPATCH && type === "MESSAGE_CREATE" && data.content) {
                let message = new Message(data, this.client);
                
                this.client.emit("message", message);
                
                if (data.content.includes(`<@${this.client.self.id}>`)) this.client.emit("mention", data);
            }

            if (op === Constants.gatewayOpCodes.INVALID_SESSION) {
                this.lastEvent = null;
                this.sessionId = null;

                if (this.client.DEBUG) logger.debug(`Invalid Session!\n${JSON.stringify(res, null, 2)}`);

                this.disconnect();
                this.connect();
            }

            if (op === Constants.gatewayOpCodes.EVENT_DISPATCH && type === "GUILD_CREATE") {
                let guild = new Guild(data, this.client);
                guild._initialize()
                this.client.guilds.set(guild.id, guild);
            }

            if (op === Constants.gatewayOpCodes.EVENT_DISPATCH && type === "CHANNEL_CREATE") {
                let channel = new GuildChannel(data, this.client);
                this.client.channels.set(channel.id, channel);
            }

            if (op === Constants.gatewayOpCodes.EVENT_DISPATCH && type === "CHANNEL_UPDATE") {
                let channel = new GuildChannel(data, this.client);
                this.client.channels.set(channel.id, channel);
            }

            if (op === Constants.gatewayOpCodes.EVENT_DISPATCH && type === "CHANNEL_DELETE") {
                let channel = new GuildChannel(data, this.client);
                this.client.channels.delete(channel.id, channel);
            }
        });

        this.ws.on("close", (code, reason) => {
            if (this.client.DEBUG) logger.debug(`Connection to Gateway Closed!\nCode: ${code}\nReason: ${reason}`);
            if (this.client.expectingClose) return;

            this.client.isConnected = false;
            this.client.ws = null;
            this.client.emit("disconnect", code)
            this.sessionId = null;
            
            if (code === 4004) {
                throw new Error("Disconnected due to an invalid token.");
                process.exit(1);
            } else if (code === 4007) {
                if (this.client.DEBUG) logger.debug("Invalid Session ID. Clearing and reconnecting.");
                
                this.lastEvent = null;
                this.sessionId = null;
                this.connect();
            } else if (code === 4011) {
                console.log("Sharding is now required. This feature is not currently implemented.");
                process.exit(1);
            } else if (code === 4008) {
                logger.warn("Disconnected due to ratelimits. Reconnecting in 30 seconds.");
                setTimeout(() => {
                    this.connect();
                }, 30000)
            } else {
                this.connect();
            }
        });

        this.ws.on("error", (error) => {
            if (this.client.DEBUG) logger.warn(error);
        });
    }

    disconnect() {
        this.expectingClose = true;
        this.ws.close();
        this.client.isConnected = false;
        this.client.ws = null;

        if (this.client.DEBUG) logger.debug("Disconnect triggered, WebSocket closing.");
    }
};

module.exports = Websocket;